from abc import ABC, abstractmethod
from calendar import timegm
from datetime import datetime
from typing import List, Dict, Union, Type, Any
from uuid import uuid4

from .errors import EventMapperNotFoundError


class Event(ABC):
    __slots__ = ('_id', '_type', '_occurred_on', '_attributes')

    def __init__(self, attributes: Dict[str, Any], **kwargs):
        self._id = kwargs.get('id', str(uuid4()))
        self._type = kwargs.get('type', 'event')
        self._occurred_on = kwargs.get('occurred_on', timegm(datetime.utcnow().utctimetuple()))
        self._attributes = attributes

    def get_meta(self) -> Dict[str, Union[str, int]]:
        return {
            'id': self._id,
            'type': self._type,
            'occurredOn': self._occurred_on,
        }

    def get_attributes(self) -> Dict[str, Any]:
        return self._attributes


class EventMapper(ABC):
    @abstractmethod
    def belongs_to(self, msg: Event) -> bool:
        pass

    @abstractmethod
    def service_name(self) -> str:
        pass

    @abstractmethod
    def name(self) -> str:
        pass

    def encode(self, msg: Event) -> Dict[str, Any]:
        attrs = self.map_attributes(msg)
        meta = msg.get_meta()
        return {
            'id': meta['id'],
            'type': meta['type'],
            'occurredOn': meta['occurredOn'],
            'attributes': attrs,
            'meta': {
                'message': f'{self.service_name()}.{self.name()}'
            },
        }

    @abstractmethod
    def decode(self, data: Dict[str, Any]) -> Event:
        pass

    @staticmethod
    def map_attributes(msg: Event) -> Dict[str, Any]:
        return msg.get_attributes()


def find_event_mapper_by_name(name: str, mappers: List[EventMapper]) -> EventMapper:
    for mapper in mappers:
        if f'{mapper.service_name()}.{mapper.name()}' == name:
            return mapper
    raise EventMapperNotFoundError.create(detail={'name': name})


def find_event_mapper_by_type(msg: Event, mappers: List[EventMapper]) -> EventMapper:
    for mapper in mappers:
        if mapper.belongs_to(msg):
            return mapper
    raise EventMapperNotFoundError.create(detail={'type': type(msg)})


class EventPublisher(ABC):
    @abstractmethod
    async def publish(self, events: List[Event]) -> None:
        pass


class EventHandler(ABC):
    @abstractmethod
    def subscribed_to(self) -> List[Type[Event]]:
        pass

    @abstractmethod
    async def handle(self, events: List[Event]) -> None:
        pass


class EventBus(ABC):
    @abstractmethod
    async def notify(self, events: List[Event]) -> None:
        pass


class SimpleEventBus(EventBus):
    _handlers: Dict[Any, List[EventHandler]]

    def __init__(self, handlers: List[EventHandler]):
        self._handlers = self._map_event_handlers(handlers)

    def add_handler(self, handler: EventHandler) -> None:
        self._handlers = self._map_event_handlers([*list(self._handlers.values()), handler])

    async def notify(self, events: List[Event]) -> None:
        for event_type, handler in self._handlers:
            for event in events:
                if isinstance(event, event_type):
                    await handler.handle([event])

    @staticmethod
    def _map_event_handlers(handlers: List[EventHandler]) -> Dict[Any, List[EventHandler]]:
        _handlers: Dict[Any, List[EventHandler]] = {}
        for handler in handlers:
            for event in handler.subscribed_to():
                if isinstance(_handlers[event], list):
                    _handlers[event].append(handler)
                else:
                    _handlers[event] = [handler]
        return _handlers



from json import dumps
from typing import Dict, Optional, Any, final
from uuid import uuid4


class BaseError(Exception):
    _id: str
    _code: str = 'code'
    _title: str = 'title'
    _detail: str
    _meta: Dict[str, Any]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args)
        self._id = kwargs.get('id', str(uuid4()))
        self._code = kwargs.get('code', self._code)
        self._title = kwargs.get('title', self._title)
        self._detail = dumps(kwargs.get('detail', {}))
        self._meta = kwargs.get('meta', {})
        if self._has_a_system_exit_exception():
            raise SystemExit(self._meta.get('exception'))

    @classmethod
    def create(cls, detail: Optional[Dict[str, Any]] = None, meta: Optional[Dict[str, Any]] = None) -> 'BaseError':
        return cls(id=str(uuid4()), detail=detail or {}, meta=meta or {})

    def _has_a_system_exit_exception(self) -> bool:
        return self._meta.get('exception_type', None) is SystemExit

    def with_exception(self, err: BaseException) -> 'BaseError':
        self._meta.update({'exception': str(err), 'exception_type': str(type(err))})
        return self

    def id(self) -> str:
        return self._id

    def code(self) -> str:
        return self._code

    def title(self) -> str:
        return self._title

    def detail(self) -> str:
        return self._detail

    def meta(self) -> Dict[str, Any]:
        return self._meta

    def __str__(self) -> str:
        return dumps({
            'id': self._id,
            'code': self._code,
            'title': self._title,
            'detail': self._detail,
            'meta': self._meta,
        }, indent=2)


class NotFoundError(BaseError):
    _code = 'not_found'
    _title = 'Not found'


class ConflictError(BaseError):
    _code = 'conflict',
    _title = 'Conflict',


class BadRequestError(BaseError):
    _code = 'bad_request',
    _title = 'Bad Request',


@final
class UnknownError(BaseError):
    _code = 'unknown',
    _title = 'Unknown error',


class IdInvalidError(ConflictError):
    _code = 'id_invalid',
    _title = 'Invalid id',


class TimestampInvalidError(ConflictError):
    _code = 'timestamp_invalid'
    _title = 'Invalid timestamp'


class DateTimeInvalidError(ConflictError):
    _code = 'datetime_invalid'
    _title = 'Invalid datetime'


class EventMapperNotFoundError(NotFoundError):
    _code = 'event_mapper_not_found'
    _title = 'Event Mapper not found'


class EventNotPublishedError(ConflictError):
    _code = 'event_not_published'
    _title = 'Event not published'


class CommandNotRegisteredError(NotFoundError):
    _code = 'command_not_registered_error'
    _title = 'Command not registered'


class QueryNotRegisteredError(NotFoundError):
    _code = 'query_not_registered_error'
    _title = 'Query not registered'



from datetime import datetime, timedelta, tzinfo
from time import time
from typing import Optional
from uuid import uuid4, UUID

from .errors import IdInvalidError, TimestampInvalidError


class Id:
    __slots__ = '_value'

    def __init__(self, value: str) -> None:
        try:
            self._value = str(UUID(value))
        except Exception as err:
            raise IdInvalidError.create(detail={'id': value}).with_exception(err)

    @classmethod
    def generate(cls) -> 'Id':
        return cls(str(uuid4()))

    @staticmethod
    def validate(value: str) -> bool:
        try:
            Id(value)
            return True
        except IdInvalidError as _:
            return False

    def value(self) -> str:
        return self._value

    def __str__(self) -> str:
        return self._value


class Timestamp:
    __slots__ = '_value'

    def __init__(self, value: float, utc: bool = True, tz: Optional[tzinfo] = None) -> None:
        try:
            if utc:
                datetime.utcfromtimestamp(value)
            else:
                datetime.fromtimestamp(value, tz)
            self._value = int(value)
        except Exception as err:
            raise TimestampInvalidError.create(detail={'timestamp': int(value)}).with_exception(err)

    @classmethod
    def now(cls) -> 'Timestamp':
        return cls(time())

    def diff(self, other: 'Timestamp', utc: bool = True, tz: Optional[tzinfo] = None) -> timedelta:
        if utc:
            return datetime.utcfromtimestamp(other.value()) - datetime.utcfromtimestamp(self._value)
        else:
            return datetime.fromtimestamp(other.value(), tz) - datetime.fromtimestamp(self._value, tz)

    def value(self) -> int:
        return self._value



from abc import ABC
from typing import List

from .cqrs import \
    Command, CommandHandler, CommandBus, SimpleCommandBus, \
    Query, Response, OptionalResponse, QueryHandler, QueryBus, SimpleQueryBus
from .errors import \
    BaseError, \
    NotFoundError, ConflictError, BadRequestError, UnknownError, \
    IdInvalidError, \
    TimestampInvalidError, DateTimeInvalidError, \
    EventMapperNotFoundError, EventNotPublishedError, \
    CommandNotRegisteredError, QueryNotRegisteredError
from .events import Event, EventMapper, EventPublisher, EventHandler, EventBus, SimpleEventBus, \
    find_event_mapper_by_name, find_event_mapper_by_type
from .value_objects import Timestamp, Id

__all__ = (
    'Aggregate',
    'AggregateRoot',
    # cqrs
    'Command',
    'CommandHandler',
    'CommandBus',
    'SimpleCommandBus',
    'Query',
    'Response',
    'OptionalResponse',
    'QueryHandler',
    'QueryBus',
    'SimpleQueryBus',
    # errors
    'BaseError',
    'NotFoundError',
    'ConflictError',
    'BadRequestError',
    'UnknownError',
    'IdInvalidError',
    'TimestampInvalidError',
    'DateTimeInvalidError',
    'EventMapperNotFoundError',
    'EventNotPublishedError',
    'CommandNotRegisteredError',
    'QueryNotRegisteredError',
    # events
    'Event',
    'EventMapper',
    'EventPublisher',
    'EventHandler',
    'EventBus',
    'SimpleEventBus',
    'find_event_mapper_by_name',
    'find_event_mapper_by_type',
    # value_objects
    'Id',
    'Timestamp'
)


class Aggregate(ABC):
    pass


class AggregateRoot(Aggregate):
    _events: List[Event]

    def __init__(self) -> None:
        self._events = []

    def pull_aggregate_events(self) -> List[Event]:
        evts = self._events
        self._events = []
        return evts

    def record_aggregate_event(self, event: Event) -> None:
        self._events.append(event)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aioddd']

package_data = \
{'': ['*']}

install_requires = \
['uuid>=1.30,<2.0']

setup_kwargs = {
    'name': 'aioddd',
    'version': '1.0.4',
    'description': 'Async Python DDD utilities library.',
    'long_description': "# Async Python DDD utilities library\n\naioddd is an async Python DDD utilities library.\n\n## Installation\n\nUse the package manager [pip](https://pip.pypa.io/en/stable/) to install aiocli.\n\n```bash\npip install aioddd\n```\n\n## Usage\n\n```python\nfrom asyncio import get_event_loop\nfrom typing import Type\nfrom aioddd import NotFoundError, \\\n    Command, CommandHandler, SimpleCommandBus, \\\n    Query, QueryHandler, OptionalResponse, SimpleQueryBus\n\n_products = []\n\nclass StoreProductCommand(Command):\n    def __init__(self, ref: str):\n        self.ref = ref\n\nclass StoreProductCommandHandler(CommandHandler):\n    def subscribed_to(self) -> Type[Command]:\n        return StoreProductCommand\n\n    async def handle(self, command: StoreProductCommand) -> None:\n        _products.append(command.ref)\n\nclass ProductNotFoundError(NotFoundError):\n    _code = 'product_not_found'\n    _title = 'Product not found'\n\nclass FindProductQuery(Query):\n    def __init__(self, ref: str):\n        self.ref = ref\n\nclass FindProductQueryHandler(QueryHandler):\n    def subscribed_to(self) -> Type[Query]:\n        return FindProductQuery\n\n    async def handle(self, query: FindProductQuery) -> OptionalResponse:\n        if query.ref != '123':\n            raise ProductNotFoundError.create(detail={'ref': query.ref})\n        return {'ref': query.ref}\n\nasync def main() -> None:\n    commands_bus = SimpleCommandBus([StoreProductCommandHandler()])\n    await commands_bus.dispatch(StoreProductCommand('123'))\n    query_bus = SimpleQueryBus([FindProductQueryHandler()])\n    response = await query_bus.ask(FindProductQuery('123'))\n    print(response)\n\n\nif __name__ == '__main__':\n    get_event_loop().run_until_complete(main())\n```\n\n## Requirements\n\n- Python >= 3.6\n\n## Contributing\nPull requests are welcome. For major changes, please open an issue first to discuss what you would like to change.\n\nPlease make sure to update tests as appropriate.\n\n## License\n[MIT](https://github.com/ticdenis/python-aioddd/blob/master/LICENSE)\n",
    'author': 'ticdenis',
    'author_email': 'denisnavarroalcaide@outlook.es',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ticdenis/python-aioddd',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
